/*
 * Copyright 2011 The Netty Project
 *
 * The Netty Project licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package io.netty.example.worldclock;

// How to generate
// ===============
// $ cd example
// $ protoc src/main/java/io/netty/example/worldclock/WorldClockProtocol.proto \
//          --java_out=src/main/java
//
// Add @SuppressWarnings("all") to the generated code not to pollute IDE task list.

option optimize_for = SPEED;

enum Continent {
  AFRICA = 0;
  AMERICA = 1;
  ANTARCTICA = 2;
  ARCTIC = 3;
  ASIA = 4;
  ATLANTIC = 5;
  AUSTRALIA = 6;
  EUROPE = 7;
  INDIAN = 8;
  MIDEAST = 9;
  PACIFIC = 10;
}

message Location {
  required Continent continent = 1;
  required string city = 2;
}

message Locations {
  repeated Location location = 1;
}

enum DayOfWeek {
  SUNDAY = 1;
  MONDAY = 2;
  TUESDAY = 3;
  WEDNESDAY = 4;
  THURSDAY = 5;
  FRIDAY = 6;
  SATURDAY = 7;
}

message LocalTime {
  required uint32 year = 1;
  required uint32 month = 2;
  required uint32 dayOfMonth = 4;
  required DayOfWeek dayOfWeek = 5;
  required uint32 hour = 6;
  required uint32 minute = 7;
  required uint32 second = 8;
}

message LocalTimes {
  repeated LocalTime localTime = 1;
}
