/*
 * Copyright 2013 The Netty Project
 *
 * The Netty Project licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package io.netty.testsuite.transport.socket;

import io.netty.bootstrap.Bootstrap;
import io.netty.buffer.Unpooled;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInboundHandlerAdapter;
import io.netty.channel.socket.SocketChannel;
import org.junit.Test;

public class WriteBeforeRegisteredTest extends AbstractClientSocketTest {

    @Test(timeout = 30000)
    public void testWriteBeforeConnect() throws Throwable {
        run();
    }

    public void testWriteBeforeConnect(Bootstrap cb) throws Throwable {
        TestHandler h = new TestHandler();
        SocketChannel ch = null;
        try {
            ch = (SocketChannel) cb.handler(h).connect().channel();
            ch.writeAndFlush(Unpooled.wrappedBuffer(new byte[] { 1 }));
        } finally {
            if (ch != null) {
                ch.close();
            }
        }
    }

    private static class TestHandler extends ChannelInboundHandlerAdapter {
        @Override
        public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception {
            cause.printStackTrace();
        }
    }
}
