/*
 * Copyright 2012 The Netty Project
 *
 * The Netty Project licenses this file to you under the Apache License,
 * version 2.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at:
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

/**
 * Encoder and decoder which transform a
 * <a href="http://code.google.com/p/protobuf/">Google Protocol Buffers</a>
 * {@link com.google.protobuf.Message} and {@link com.google.protobuf.nano.MessageNano} into a
 * {@link io.netty.buffer.ByteBuf} and vice versa.
 */
package io.netty.handler.codec.protobuf;
